@@ -41,7 +41,7 @@
 #include "soup-socket.h"
 #include "soup-ssl.h"
 
-GHashTable *soup_servers;  /* KEY: hostname, VALUE: SoupServer */
+GHashTable *soup_hosts;  /* KEY: hostname, VALUE: SoupHost */
 
 static gint connection_count = 0;
 
@@ -85,17 +85,31 @@ static guint
 soup_context_uri_hash (gconstpointer key)
 {
 	const SoupUri *uri = key;
-	guint ret = 0;
+	guint ret;
 
-	ret += uri->protocol;
-	ret += g_str_hash (uri->path ? uri->path : "");
-	ret += g_str_hash (uri->querystring ? uri->querystring : "");
-	ret += g_str_hash (uri->user ? uri->user : "");
-	ret += g_str_hash (uri->passwd ? uri->passwd : "");
+	ret = uri->protocol;
+	if (uri->path)
+		ret += g_str_hash (uri->path);
+	if (uri->querystring)
+		ret += g_str_hash (uri->querystring);
+	if (uri->user)
+		ret += g_str_hash (uri->user);
+	if (uri->passwd)
+		ret += g_str_hash (uri->passwd);
 
 	return ret;
 }
 
+static inline gboolean
+parts_equal (const char *one, const char *two)
+{
+	if (!one && !two)
+		return TRUE;
+	if (!one || !two)
+		return FALSE;
+	return !strcmp (one, two);
+}
+
 /**
  * soup_context_uri_equal:
  * @v1: a %SoupUri
@@ -110,18 +124,18 @@ soup_context_uri_equal (gconstpointer v1, gconstpointer v2)
 	const SoupUri *one = v1;
 	const SoupUri *two = v2;
 
-	if (one->protocol == two->protocol &&
-	    !strcmp (one->path ? one->path : "",
-		     two->path ? two->path : "") &&
-	    !strcmp (one->querystring ? one->querystring : "",
-		     two->querystring ? two->querystring : "") &&
-	    !strcmp (one->user ? one->user : "",
-		     two->user ? two->user : "") &&
-	    !strcmp (one->passwd ? one->passwd : "",
-		     two->passwd ? two->passwd : ""))
-		return TRUE;
+	if (one->protocol != two->protocol)
+		return FALSE;
+	if (!parts_equal (one->path, two->path))
+		return FALSE;
+	if (!parts_equal (one->user, two->user))
+		return FALSE;
+	if (!parts_equal (one->passwd, two->passwd))
+		return FALSE;
+	if (!parts_equal (one->querystring, two->querystring))
+		return FALSE;
 
-	return FALSE;
+	return TRUE;
 }
 
 /**
@@ -137,22 +151,22 @@ soup_context_uri_equal (gconstpointer v1, gconstpointer v2)
 SoupContext *
 soup_context_from_uri (SoupUri *suri)
 {
-	SoupServer *serv = NULL;
+	SoupHost *serv = NULL;
 	SoupContext *ret = NULL;
 
 	g_return_val_if_fail (suri != NULL, NULL);
 	g_return_val_if_fail (suri->protocol != 0, NULL);
 
-	if (!soup_servers)
-		soup_servers = g_hash_table_new (soup_str_case_hash,
-						 soup_str_case_equal);
+	if (!soup_hosts)
+		soup_hosts = g_hash_table_new (soup_str_case_hash,
+					       soup_str_case_equal);
 	else
-		serv = g_hash_table_lookup (soup_servers, suri->host);
+		serv = g_hash_table_lookup (soup_hosts, suri->host);
 
 	if (!serv) {
-		serv = g_new0 (SoupServer, 1);
+		serv = g_new0 (SoupHost, 1);
 		serv->host = g_strdup (suri->host);
-		g_hash_table_insert (soup_servers, serv->host, serv);
+		g_hash_table_insert (soup_hosts, serv->host, serv);
 	}
 
 	if (!serv->contexts)
@@ -189,6 +203,15 @@ soup_context_ref (SoupContext *ctx)
 	ctx->refcnt++;
 }
 
+static gboolean
+remove_auth (gchar *path, SoupAuth *auth)
+{
+	g_free (path);
+	soup_auth_free (auth);
+
+	return TRUE;
+}
+
 /**
  * soup_context_unref:
  * @ctx: a %SoupContext.
@@ -205,36 +228,66 @@ soup_context_unref (SoupContext *ctx)
 	--ctx->refcnt;
 
 	if (ctx->refcnt == 0) {
-		SoupServer *serv = ctx->server;
+		SoupHost *serv = ctx->server;
 
 		g_hash_table_remove (serv->contexts, ctx->uri);
 
 		if (g_hash_table_size (serv->contexts) == 0) {
-			GSList *conns = serv->connections;
+			/*
+			 * Remove this host from the active hosts hash
+			 */
+			g_hash_table_remove (soup_hosts, serv->host);
 
-			g_hash_table_remove (soup_servers, serv->host);
-
-			while (conns) {
-				SoupConnection *conn = conns->data;
-				soup_socket_unref (conn->socket);
-				g_free (conn);
-				connection_count--;
-
-				conns = conns->next;
+			/* 
+			 * Free all cached SoupAuths
+			 */
+			if (serv->valid_auths) {
+				g_hash_table_foreach_remove (
+					serv->valid_auths,
+					(GHRFunc) remove_auth,
+					NULL);
+				g_hash_table_destroy (serv->valid_auths);
 			}
 
-			g_free (serv->host);
-			g_slist_free (serv->connections);
 			g_hash_table_destroy (serv->contexts);
+			g_free (serv->host);
 			g_free (serv);
 		}
 
-		if (ctx->auth) soup_auth_free (ctx->auth);
 		soup_uri_free (ctx->uri);
 		g_free (ctx);
 	}
 }
 
+static void
+connection_free (SoupConnection *conn)
+{
+	g_return_if_fail (conn != NULL);
+
+	conn->server->connections =
+		g_slist_remove (conn->server->connections, conn);
+
+	soup_context_unref (conn->context);
+	soup_socket_unref (conn->socket);
+	g_source_remove (conn->death_tag);
+	g_free (conn);
+
+	connection_count--;
+}
+
+static gboolean 
+connection_death (GIOChannel*     iochannel,
+		  GIOCondition    condition,
+		  SoupConnection *conn)
+{
+	if (!conn->in_use) {
+		connection_free (conn);
+		return FALSE;
+	}
+
+	return TRUE;
+}
+
 struct SoupConnectData {
 	SoupContext           *ctx;
 	SoupConnectCallbackFn  cb;
@@ -244,6 +297,44 @@ struct SoupConnectData {
 	gpointer               connect_tag;
 };
 
+static void
+prune_connection_foreach (gchar               *hostname,
+			  SoupHost            *serv,
+			  SoupConnection     **last)
+{
+	GSList *conns = serv->connections;
+
+	while (conns) {
+		SoupConnection *conn = conns->data;
+
+		if (!conn->in_use) {
+			if (*last == NULL ||
+			    (*last)->last_used_id > conn->last_used_id)
+				*last = conn;
+		}
+
+		conns = conns->next;
+	}
+}
+
+static gboolean
+prune_least_used_connection (void)
+{
+	SoupConnection *last = NULL;
+
+	g_hash_table_foreach (soup_hosts, 
+			      (GHFunc) prune_connection_foreach, 
+			      &last);
+	if (last) {
+		connection_free (last);
+		return TRUE;
+	}
+
+	return FALSE;
+}
+
+static gboolean retry_connect_timeout_cb (struct SoupConnectData *data);
+
 static void
 soup_context_connect_cb (SoupSocket              *socket,
 			 SoupSocketConnectStatus  status,
@@ -251,135 +342,154 @@ soup_context_connect_cb (SoupSocket              *socket,
 {
 	struct SoupConnectData *data = user_data;
 	SoupContext            *ctx = data->ctx;
-	SoupConnectCallbackFn   cb = data->cb;
-	gpointer                cb_data = data->user_data;
 	SoupConnection         *new_conn;
-
-	g_free (data);
+	GIOChannel             *chan;
 
 	switch (status) {
 	case SOUP_SOCKET_CONNECT_ERROR_NONE:
 		new_conn = g_new0 (SoupConnection, 1);
 		new_conn->server = ctx->server;
-		new_conn->context = ctx;
 		new_conn->socket = socket;
 		new_conn->port = ctx->uri->port;
 		new_conn->keep_alive = TRUE;
 		new_conn->in_use = TRUE;
 		new_conn->last_used_id = 0;
 
-		ctx->server->connections =
-			g_slist_prepend (ctx->server->connections,
-					 new_conn);
+		new_conn->context = ctx;
+		soup_context_ref (ctx);
 
-		(*cb) (ctx, SOUP_CONNECT_ERROR_NONE, new_conn, cb_data);
+		chan = soup_connection_get_iochannel (new_conn);
+		new_conn->death_tag = 
+			g_io_add_watch (chan,
+					G_IO_ERR | G_IO_HUP | G_IO_NVAL,
+					(GIOFunc) connection_death,
+					new_conn);
+		g_io_channel_unref (chan);
+
+		ctx->server->connections =
+			g_slist_prepend (ctx->server->connections, new_conn);
+
+		(*data->cb) (ctx, 
+			     SOUP_CONNECT_ERROR_NONE, 
+			     new_conn, 
+			     data->user_data);
 		break;
 	case SOUP_SOCKET_CONNECT_ERROR_ADDR_RESOLVE:
 		connection_count--;
 
-		(*cb) (ctx, SOUP_CONNECT_ERROR_ADDR_RESOLVE, NULL, cb_data);
+		(*data->cb) (ctx, 
+			     SOUP_CONNECT_ERROR_ADDR_RESOLVE, 
+			     NULL, 
+			     data->user_data);
 		break;
 	case SOUP_SOCKET_CONNECT_ERROR_NETWORK:
 		connection_count--;
 
-		(*cb) (ctx, SOUP_CONNECT_ERROR_NETWORK, NULL, cb_data);
+		/*
+		 * Check if another connection exists to this server
+		 * before reporting error. 
+		 */
+		if (ctx->server->connections) {
+			data->timeout_tag =
+				g_timeout_add (
+					150,
+					(GSourceFunc) retry_connect_timeout_cb,
+					data);
+			return;
+		}
+
+		(*data->cb) (ctx, 
+			     SOUP_CONNECT_ERROR_NETWORK, 
+			     NULL, 
+			     data->user_data);
 		break;
 	}
+
+	soup_context_unref (ctx);
+	g_free (data);
 }
 
-static SoupConnection *
-soup_try_existing_connections (SoupContext *ctx)
+static gboolean
+try_existing_connections (SoupContext           *ctx,
+			  SoupConnectCallbackFn  cb,
+			  gpointer               user_data)
 {
 	GSList *conns = ctx->server->connections;
-
-	/* 
-	 * FIXME FIXME FIXME: This stuff isn't working quite right yet.
-	 * It sometimes throws errors when a connection gets HUP'ed, so
-	 * we'll just disable it for now.
-	 * 
-	 * Seems fixed now - orph 8/28/01.
-	 */
-	/*
-	 * return NULL;
-	 */
 	
 	while (conns) {
 		SoupConnection *conn = conns->data;
 
-		if (!conn->in_use &&
-		    conn->port == ctx->uri->port &&
-		    conn->keep_alive)
-			return conn;
+		if (conn->in_use == FALSE &&
+		    conn->keep_alive == TRUE &&
+		    conn->port == (guint) ctx->uri->port) {
+			/* Set connection to in use */
+			conn->in_use = TRUE;
+
+			/* Reset connection context */
+			soup_context_ref (ctx);
+			soup_context_unref (conn->context);
+			conn->context = ctx;
+
+			/* Issue success callback */
+			(*cb) (ctx, SOUP_CONNECT_ERROR_NONE, conn, user_data);
+			return TRUE;
+		}
 
 		conns = conns->next;
 	}
 
-	return NULL;
-}
-
-struct SoupConnDesc {
-	SoupServer     *serv;
-	SoupConnection *conn;
-};
-
-static void
-soup_prune_foreach (gchar *hostname,
-		    SoupServer *serv,
-		    struct SoupConnDesc *last)
-{
-	GSList *conns = serv->connections;
-
-	while (conns) {
-		SoupConnection *conn = conns->data;
-		if (!conn->in_use)
-			if (last->conn == NULL ||
-			    last->conn->last_used_id > conn->last_used_id) {
-				last->conn = conn;
-				last->serv = serv;
-			}
-
-		conns = conns->next;
-	}
-}
-
-static gboolean
-soup_prune_least_used_connection (void)
-{
-	struct SoupConnDesc last;
-	last.serv = NULL;
-	last.conn = NULL;
-
-	g_hash_table_foreach (soup_servers, (GHFunc) soup_prune_foreach, &last);
-
-	if (last.conn) {
-		last.serv->connections =
-			g_slist_remove (last.serv->connections, last.conn);
-		soup_socket_unref (last.conn->socket);
-		g_free (last.conn);
-
-		connection_count--;
-
-		return TRUE;
-	}
-
 	return FALSE;
 }
 
 static gboolean
-soup_prune_timeout (struct SoupConnectData *data)
+try_create_connection (struct SoupConnectData **dataptr)
 {
-	guint conn_limit = soup_get_connection_limit();
+	struct SoupConnectData *data = *dataptr;
+	gint conn_limit = soup_get_connection_limit ();
+	gpointer connect_tag;
 
+	/* 
+	 * Check if we are allowed to create a new connection, otherwise wait
+	 * for next timeout.  
+	 */
 	if (conn_limit &&
 	    connection_count >= conn_limit &&
-	    !soup_try_existing_connections (data->ctx) &&
-	    !soup_prune_least_used_connection ())
-		return TRUE;
+	    !prune_least_used_connection ()) {
+		data->connect_tag = 0;
+		return FALSE;
+	}
 
-	soup_context_get_connection (data->ctx, data->cb, data->user_data);
-	g_free (data);
+	connection_count++;
 
-	return FALSE;
+	data->timeout_tag = 0;
+	connect_tag = soup_socket_connect (data->ctx->uri->host,
+					   data->ctx->uri->port,
+					   soup_context_connect_cb,
+					   data);
+	/* 
+	 * NOTE: soup_socket_connect can fail immediately and call our
+	 * callback which will delete the state.  
+	 */
+	if (connect_tag)
+		data->connect_tag = connect_tag;
+	else
+		*dataptr = NULL;
+
+	return TRUE;
+}
+
+static gboolean
+retry_connect_timeout_cb (struct SoupConnectData *data)
+{
+	if (try_existing_connections (data->ctx, 
+				      data->cb, 
+				      data->user_data)) {
+		soup_context_unref (data->ctx);
+		g_free (data);
+		return FALSE;
+	}
+
+	return try_create_connection (&data) == FALSE;
 }
 
 /**
@@ -408,41 +518,26 @@ soup_context_get_connection (SoupContext           *ctx,
 			     SoupConnectCallbackFn  cb,
 			     gpointer               user_data)
 {
-	SoupConnection *conn;
 	struct SoupConnectData *data;
-	guint conn_limit;
 
 	g_return_val_if_fail (ctx != NULL, NULL);
 
-	if ((conn = soup_try_existing_connections (ctx))) {
-		conn->in_use = TRUE;
-		(*cb) (ctx, SOUP_CONNECT_ERROR_NONE, conn, user_data);
+	/* Look for an existing unused connection */
+	if (try_existing_connections (ctx, cb, user_data))
 		return NULL;
-	}
 
 	data = g_new0 (struct SoupConnectData, 1);
-	data->ctx = ctx;
 	data->cb = cb;
 	data->user_data = user_data;
 
-	conn_limit = soup_get_connection_limit ();
+	data->ctx = ctx;
+	soup_context_ref (ctx);
 
-	if (conn_limit &&
-	    connection_count >= conn_limit &&
-	    !soup_prune_least_used_connection ())
+	if (!try_create_connection (&data))
 		data->timeout_tag =
-			g_timeout_add (500,
-				       (GSourceFunc) soup_prune_timeout,
+			g_timeout_add (150,
+				       (GSourceFunc) retry_connect_timeout_cb,
 				       data);
-	else {
-		connection_count++;
-
-		data->connect_tag =
-			soup_socket_connect (ctx->uri->host,
-					     ctx->uri->port,
-					     soup_context_connect_cb,
-					     data);
-	}
 
 	return data;
 }
@@ -464,8 +559,10 @@ soup_context_cancel_connect (SoupConnectId tag)
 
 	if (data->timeout_tag)
 		g_source_remove (data->timeout_tag);
-	else if (data->connect_tag)
+	else if (data->connect_tag) {
+		connection_count--;
 		soup_socket_connect_cancel (data->connect_tag);
+	}
 
 	g_free (data);
 }
@@ -501,14 +598,9 @@ soup_connection_release (SoupConnection *conn)
 
 	if (conn->keep_alive) {
 		conn->last_used_id = ++most_recently_used_id;
-		conn->in_use = FALSE;
-	} else {
-		conn->server->connections =
-			g_slist_remove (conn->server->connections, conn);
-		soup_socket_unref (conn->socket);
-		g_free (conn);
-		connection_count--;
-	}
+		conn->in_use = FALSE;		
+	} else
+		connection_free (conn);
 }
 
 static void
@@ -552,7 +644,7 @@ soup_connection_get_iochannel (SoupConnection *conn)
 }
 
 /**
- * soup_connection_set_keepalive:
+ * soup_connection_set_keep_alive:
  * @conn: a %SoupConnection.
  * @keep_alive: boolean keep-alive value.
  *
@@ -566,7 +658,7 @@ soup_connection_set_keep_alive (SoupConnection *conn, gboolean keep_alive)
 }
 
 /**
- * soup_connection_set_keepalive:
+ * soup_connection_is_keep_alive:
  * @conn: a %SoupConnection.
  *
  * Returns the keep-alive flag for the %SoupConnection pointed to by
@@ -587,14 +679,16 @@ soup_connection_is_keep_alive (SoupConnection *conn)
  * soup_connection_get_context:
  * @conn: a %SoupConnection.
  *
- * Returns the %SoupContext from which @conn was created.
+ * Returns the %SoupContext from which @conn was created, with an added ref.
  *
- * Return value: the %SoupContext associated with @conn.
+ * Return value: the %SoupContext associated with @conn.  Unref when finished.
  */
 SoupContext *
 soup_connection_get_context (SoupConnection *conn)
 {
 	g_return_val_if_fail (conn != NULL, FALSE);
+	
+	soup_context_ref (conn->context);
 	return conn->context;
 }
 
